/*
 * Code to loop a framebuffer to GPIO pins
 * Copyright (c) 2013 Sprite_tm <jeroen@spritesmods.com>
 * Heavily hacked from the original source:
 * servod.c Multiple Servo Driver for the RaspberryPi
 * Copyright (c) 2013 Richard Hirst <richardghirst@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <stdint.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#define PAGE_SIZE		4096

#define DMA_BASE		0x20007000
#define DMA_LEN			0x24
#define GPIO_BASE		0x20200000
#define GPIO_LEN		0x100

#define DMA_NO_WIDE_BURSTS	(1<<26)
#define DMA_END			(1<<1)
#define DMA_INT			(1<<2)
#define DMA_WAIT_RESP		(1<<3)
#define DMA_DEST_INC		(1<<4)
#define DMA_D_DREQ		(1<<6)
#define DMA_SRC_INC		(1<<8)
#define DMA_PER_MAP(x)		((x)<<16)
#define DMA_WAITS(x)		((x)<<21)
#define DMA_RESET		(1<<31)
#define DMA_TDMODE		(1<<1)

#define DMA_CS			(0x00/4)
#define DMA_CONBLK_AD		(0x04/4)
#define DMA_DEBUG		(0x20/4)

#define GPIO_FSEL0		(0x00/4)
#define GPIO_SET0		(0x1c/4)
#define GPIO_CLR0		(0x28/4)
#define GPIO_LEV0		(0x34/4)
#define GPIO_PULLEN		(0x94/4)
#define GPIO_PULLCLK		(0x98/4)

#define GPIO_MODE_IN		0
#define GPIO_MODE_OUT		1

#define NOWAITS 0

//We control three sub-screen of 224x16 pixels each sample we send. We do
//4-bit pwm, so we send 16 sub-frames per frame. Each sample is sent twice,
//once with low and once with high clock-signal.
#define TOTAL_SAMPLES (224*16*16*2)

//Because the page mapping between virtual and physical addresses doesn't have
//to be continuous, we can only upload one page full of continuous memory,
//aka 256 gpio_dma_t samples.
#define NUM_SAMPLES 256
#define NUM_CBS (TOTAL_SAMPLES/NUM_SAMPLES)

//A DMA control block
typedef struct {
	uint32_t info, src, dst, length,
		 stride, next, pad[2];
} dma_cb_t;


//The GPIO register structure
typedef struct {
	uint32_t gpioset0;
	uint32_t gpioset1;
	uint32_t rsv;
	uint32_t gpioclr0;
} gpio_dma_t;

//Data struct to store everything in
struct ctl {
	gpio_dma_t sample[NUM_CBS][NUM_SAMPLES];
	dma_cb_t cb[NUM_CBS];
};

#define NUM_PAGES		((sizeof(struct ctl) + PAGE_SIZE - 1) / PAGE_SIZE)

typedef struct {
	uint8_t *virtaddr;
	uint32_t physaddr;
} page_map_t;

page_map_t *page_map;

static uint8_t *virtbase;

static volatile uint32_t *dma_reg;
static volatile uint32_t *gpio_reg;

static int gpio_mask;

static void terminate(int dummy) {
	dma_reg[DMA_CS] = DMA_RESET;
	exit(1);
}

static void gpio_set_mode(uint32_t pin, uint32_t mode) {
	uint32_t fsel = gpio_reg[GPIO_FSEL0 + pin/10];

	fsel &= ~(7 << ((pin % 10) * 3));
	fsel |= mode << ((pin % 10) * 3);
	gpio_reg[GPIO_FSEL0 + pin/10] = fsel;
}

static void gpio_set(int pin, int level) {
	if (level)
		gpio_reg[GPIO_SET0] = 1 << pin;
	else
		gpio_reg[GPIO_CLR0] = 1 << pin;
}

static void fatal(char *fmt, ...) {
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	terminate(0);
}

static uint32_t mem_virt_to_phys(void *virt) {
	uint32_t offset = (uint8_t *)virt - virtbase;
	return page_map[offset / PAGE_SIZE].physaddr + (offset % PAGE_SIZE);
}

static void *map_peripheral(uint32_t base, uint32_t len) {
	int fd = open("/dev/mem", O_RDWR);
	void * vaddr;

	if (fd < 0) fatal("servod: Failed to open /dev/mem: %m\n");
	vaddr = mmap(NULL, len, PROT_READ|PROT_WRITE, MAP_SHARED, fd, base);
	if (vaddr == MAP_FAILED) fatal("servod: Failed to map peripheral at 0x%08x: %m\n", base);
	close(fd);

	return vaddr;
}

static void make_pagemap(void) {
	int i, fd, memfd, pid;
	char pagemap_fn[64];

	page_map = malloc(NUM_PAGES * sizeof(*page_map));
	if (page_map == 0)
		fatal("servod: Failed to malloc page_map: %m\n");
	memfd = open("/dev/mem", O_RDWR);
	if (memfd < 0)
		fatal("servod: Failed to open /dev/mem: %m\n");
	pid = getpid();
	sprintf(pagemap_fn, "/proc/%d/pagemap", pid);
	fd = open(pagemap_fn, O_RDONLY);
	if (fd < 0)
		fatal("servod: Failed to open %s: %m\n", pagemap_fn);
	if (lseek(fd, (uint32_t)virtbase >> 9, SEEK_SET) !=
						(uint32_t)virtbase >> 9) {
		fatal("servod: Failed to seek on %s: %m\n", pagemap_fn);
	}
	for (i = 0; i < NUM_PAGES; i++) {
		uint64_t pfn;
		page_map[i].virtaddr = virtbase + i * PAGE_SIZE;
		// Following line forces page to be allocated
		page_map[i].virtaddr[0] = 0;
		if (read(fd, &pfn, sizeof(pfn)) != sizeof(pfn))
			fatal("servod: Failed to read %s: %m\n", pagemap_fn);
		if (((pfn >> 55) & 0x1bf) != 0x10c)
			fatal("servod: Page %d not present (pfn 0x%016llx)\n", i, pfn);
		page_map[i].physaddr = ((uint32_t)pfn * PAGE_SIZE) | 0x40000000;
	}
	close(fd);
	close(memfd);
}

static void setup_sighandlers(void) {
	int i;

	// Catch all signals possible - it is vital we kill the DMA engine
	// on process exit!
	for (i = 0; i < 64; i++) {
		struct sigaction sa;

		memset(&sa, 0, sizeof(sa));
		sa.sa_handler = terminate;
		sigaction(i, &sa, NULL);
	}
}

static void init_ctrl_data(void) {
	int x;
	struct ctl *ctl = (struct ctl *)virtbase;
	dma_cb_t *cbp = ctl->cb;
	uint32_t phys_gpset0 = 0x7e200000 + 0x1C;

	//Fill in the DMA control blocks.
	for (x=0; x<NUM_CBS; x++) {
		cbp[x].info = DMA_NO_WIDE_BURSTS | DMA_WAITS(NOWAITS) | DMA_SRC_INC | DMA_DEST_INC | DMA_TDMODE;
		cbp[x].src = mem_virt_to_phys(&ctl->sample[x][0]);
		cbp[x].dst = phys_gpset0;
		cbp[x].length = ((NUM_SAMPLES-1)<<16)|sizeof(gpio_dma_t);
		cbp[x].stride = (-sizeof(gpio_dma_t))<<16;
		cbp[x].next = mem_virt_to_phys(&ctl->cb[x+1]); //point to next control block
//		cbp[x].next = mem_virt_to_phys(&ctl->cb[x]);
	}
	//Final control block -> loop back to first
	cbp[NUM_CBS-1].next = mem_virt_to_phys(&ctl->cb[0]);
}

static void init_hardware(void) {
	struct ctl *ctl = (struct ctl *)virtbase;

	// Initialise the DMA engine
	dma_reg[DMA_CS] = DMA_RESET;
	usleep(10);
	dma_reg[DMA_CS] = DMA_INT | DMA_END;
	dma_reg[DMA_CONBLK_AD] = mem_virt_to_phys(&ctl->cb[0]);
	dma_reg[DMA_DEBUG] = 7; // clear debug error flags
	dma_reg[DMA_CS] = 0x10880001;	// go, mid priority, wait for outstanding writes

}


void dmaloopInit(int gpiomask) {
	int i;
	struct ctl *ctl;
	gpio_mask=gpiomask;

	setup_sighandlers();

	dma_reg = map_peripheral(DMA_BASE, DMA_LEN);
	gpio_reg = map_peripheral(GPIO_BASE, GPIO_LEN);

	virtbase = mmap(NULL, NUM_PAGES * PAGE_SIZE, PROT_READ|PROT_WRITE,
			MAP_SHARED|MAP_ANONYMOUS|MAP_NORESERVE|MAP_LOCKED,
			-1, 0);
	if (virtbase == MAP_FAILED)
		fatal("servod: Failed to mmap physical pages: %m\n");
	if ((unsigned long)virtbase & (PAGE_SIZE-1))
		fatal("servod: Virtual address is not page aligned\n");
	memset(virtbase, 0, NUM_PAGES * PAGE_SIZE);
	ctl = (struct ctl *)virtbase;

	make_pagemap();

	init_ctrl_data();
	init_hardware();

	//Set GPIO pins
	for (i=0; i<32; i++) {
		if (gpiomask&(1<<i)) gpio_set_mode(i, GPIO_MODE_OUT);
	}

}

void dmaloopWriteWord(int pos, int val) {
	struct ctl *ctl = (struct ctl *)virtbase;
	if (pos>NUM_SAMPLES*NUM_CBS) {
		printf("Aiee! Write outside of sample mem: %i\n", pos);
	}
	ctl->sample[pos/NUM_SAMPLES][pos%NUM_SAMPLES].gpioset0=val&gpio_mask;
	ctl->sample[pos/NUM_SAMPLES][pos%NUM_SAMPLES].gpioclr0=(~val)&gpio_mask;
}

